class WordNetError(Exception):
    def __init__(self, msg):
        super(WordNetError, self).__init__(msg)


class SynsetError(Exception):
    def __init__(self, msg):
        super(SynsetError, self).__init__(msg)





















